//---------------------------------------------------------------------------
//
// Copyright (c) 2008-2016 AlazarTech, Inc.
//
// AlazarTech, Inc. licenses this software under specific terms and conditions.
// Use of any of the software or derivatives thereof in any product without an
// AlazarTech digitizer board is strictly prohibited.
//
// AlazarTech, Inc. provides this software AS IS, WITHOUT ANY WARRANTY, EXPRESS
// OR IMPLIED, INCLUDING, WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY OR
// FITNESS FOR A PARTICULAR PURPOSE. AlazarTech makes no guarantee or
// representations regarding the use of, or the results of the use of, the
// software and documentation in terms of correctness, accuracy, reliability,
// currentness, or otherwise; and you rely on the software, documentation and
// results solely at your own risk.
//
// IN NO EVENT SHALL ALAZARTECH BE LIABLE FOR ANY LOSS OF USE, LOSS OF BUSINESS,
// LOSS OF PROFITS, INDIRECT, INCIDENTAL, SPECIAL OR CONSEQUENTIAL DAMAGES OF
// ANY KIND. IN NO EVENT SHALL ALAZARTECH'S TOTAL LIABILITY EXCEED THE SUM PAID
// TO ALAZARTECH FOR THE PRODUCT LICENSED HEREUNDER.
//
//---------------------------------------------------------------------------

// AcqToDisk.cpp :
//
// This program demonstrates how to configure an ATS9373 to make a No-PreTrigger
// (NPT) AutoDMA acquisition into a large user-space buffer. It divides the user
// space buffer into segments, and makes DMA transfers into the buffer segments.
//

#include "AlazarApi.h"
#include "AlazarCmd.h"
#include "AlazarError.h"
#include <conio.h>
#include <stdio.h>

// Globals variables

U16 *Buffer = NULL;

// Forward declarations

BOOL ConfigureBoard(HANDLE boardHandle);
BOOL AcquireData(HANDLE boardHandle);

//----------------------------------------------------------------------------
//
// Function    :  main
//
// Description :  Program entry point
//
//----------------------------------------------------------------------------

int main(int argc, char *argv[]) {
  // TODO: Select a board

  U32 systemId = 1;
  U32 boardId = 1;

  // Get a handle to the board

  HANDLE boardHandle = AlazarGetBoardBySystemID(systemId, boardId);
  if (boardHandle == NULL) {
    printf("Error: Unable to open board system Id %u board Id %u\n", systemId,
           boardId);
    return 1;
  }

  // Configure the board's sample rate, input, and trigger settings

  if (!ConfigureBoard(boardHandle)) {
    printf("Error: Configure board failed\n");
    return 1;
  }

  // Make an acquisition, optionally saving sample data to a file

  if (!AcquireData(boardHandle)) {
    printf("Error: Acquisition failed\n");
    return 1;
  }

  return 0;
}

//----------------------------------------------------------------------------
//
// Function    :  ConfigureBoard
//
// Description :  Configure sample rate, input, and trigger settings
//
//----------------------------------------------------------------------------

BOOL ConfigureBoard(HANDLE boardHandle) {
  RETURN_CODE retCode;

  // TODO: Specify the sample rate (see sample rate id below)

  double samplesPerSec = 2.e9;

  // TODO: Select clock parameters as required to generate this sample rate.

  retCode = AlazarSetCaptureClock(boardHandle,         // HANDLE -- board handle
                                  INTERNAL_CLOCK,      // U32 -- clock source id
                                  SAMPLE_RATE_2000MSPS, // U32 -- sample rate id
                                  CLOCK_EDGE_RISING,   // U32 -- clock edge id
                                  0 // U32 -- clock decimation
                                  );
  if (retCode != ApiSuccess) {
    printf("Error: AlazarSetCaptureClock failed -- %s\n",
           AlazarErrorToText(retCode));
    return FALSE;
  }

  // TODO: Select CHA input parameters as required

  retCode = AlazarInputControl(boardHandle, // HANDLE -- board handle
                               CHANNEL_A,   // U8 -- input channel
                               DC_COUPLING, // U32 -- input coupling id
                               INPUT_RANGE_PM_400_MV, // U32 -- input range id
                               IMPEDANCE_50_OHM // U32 -- input impedance id
                               );
  if (retCode != ApiSuccess) {
    printf("Error: AlazarInputControl failed -- %s\n",
           AlazarErrorToText(retCode));
    return FALSE;
  }

  // TODO: Select CHB input parameters as required

  retCode = AlazarInputControl(boardHandle, // HANDLE -- board handle
                               CHANNEL_B,   // U8 -- channel identifier
                               DC_COUPLING, // U32 -- input coupling id
                               INPUT_RANGE_PM_400_MV, // U32 -- input range id
                               IMPEDANCE_50_OHM // U32 -- input impedance id
                               );
  if (retCode != ApiSuccess) {
    printf("Error: AlazarInputControl failed -- %s\n",
           AlazarErrorToText(retCode));
    return FALSE;
  }

  // TODO: Select trigger inputs and levels as required

  retCode = AlazarSetTriggerOperation(
      boardHandle,            // HANDLE -- board handle
      TRIG_ENGINE_OP_J,       // U32 -- trigger operation
      TRIG_ENGINE_J,          // U32 -- trigger engine id
      TRIG_EXTERNAL,          // U32 -- trigger source id
      TRIGGER_SLOPE_POSITIVE, // U32 -- trigger slope id
      160,           // U32 -- trigger level from 0 (-range) to 255 (+range)
      TRIG_ENGINE_K, // U32 -- trigger engine id
      TRIG_DISABLE,  // U32 -- trigger source id for engine K
      TRIGGER_SLOPE_POSITIVE, // U32 -- trigger slope id
      128 // U32 -- trigger level from 0 (-range) to 255 (+range)
      );
  if (retCode != ApiSuccess) {
    printf("Error: AlazarSetTriggerOperation failed -- %s\n",
           AlazarErrorToText(retCode));
    return FALSE;
  }

  // TODO: Select external trigger parameters as required

  retCode = AlazarSetExternalTrigger(
      boardHandle, // HANDLE -- board handle
      DC_COUPLING, // U32 -- external trigger coupling id
      ETR_5V       // U32 -- external trigger range id
      );

  // TODO: Set trigger delay as required.

  double triggerDelay_sec = 0.;
  U32 triggerDelay_samples = (U32)(triggerDelay_sec * samplesPerSec + 0.5);
  retCode = AlazarSetTriggerDelay(boardHandle, triggerDelay_samples);
  if (retCode != ApiSuccess) {
    printf("Error: AlazarSetTriggerDelay failed -- %s\n",
           AlazarErrorToText(retCode));
    return FALSE;
  }

  // TODO: Set trigger timeout as required.

  // NOTE:
  // The board will wait for a for this amount of time for a trigger event.
  // If a trigger event does not arrive, then the board will automatically
  // trigger. Set the trigger timeout value to 0 to force the board to wait
  // forever for a trigger event.
  //
  // IMPORTANT:
  // The trigger timeout value should be set to zero after appropriate
  // trigger parameters have been determined, otherwise the
  // board may trigger if the timeout interval expires before a
  // hardware trigger event arrives.

  double triggerTimeout_sec = 0;
  U32 triggerTimeout_clocks = (U32)(triggerTimeout_sec / 10.e-6 + 0.5);

  retCode = AlazarSetTriggerTimeOut(boardHandle, // HANDLE -- board handle
                                    triggerTimeout_clocks // U32 -- timeout_sec
                                                          // / 10.e-6 (0 means
                                                          // wait forever)
                                    );
  if (retCode != ApiSuccess) {
    printf("Error: AlazarSetTriggerTimeOut failed -- %s\n",
           AlazarErrorToText(retCode));
    return FALSE;
  }

  // TODO: Configure AUX I/O connector as required

  retCode = AlazarConfigureAuxIO(boardHandle,     // HANDLE -- board handle
                                 AUX_OUT_TRIGGER, // U32 -- mode
                                 0                // U32 -- parameter
                                 );
  if (retCode != ApiSuccess) {
    printf("Error: AlazarConfigureAuxIO failed -- %s\n",
           AlazarErrorToText(retCode));
    return FALSE;
  }

  return TRUE;
}

//----------------------------------------------------------------------------
//
// Function    :  AcquireData
//
// Description :  Perform an acquisition, optionally saving data to file.
//
//----------------------------------------------------------------------------

BOOL AcquireData(HANDLE boardHandle) {

  // The DMA buffers need to be page-aligned (a page is 4096 bytes). Since the
  // user buffer will be split in a succession of DMA buffers, the DMA buffers
  // need to be a multiple of a page size.

  // TODO: Select the number of post-trigger samples per record

  U32 samplesPerRecord = 20480;

  // TODO: Specify the total number of records to capture

  U32 recordsPerUserBuffer = 40960;

  // TODO: Specify the number of records in each DMA transfer. This must divide
  // evenly recordsPerUserBuffer.

  U32 recordsPerDmaBuffer = 16;

  // TODO: Select which channels to capture (A, B, or all)

  U32 channelMask = CHANNEL_A | CHANNEL_B;

  // TODO: Select if you wish to save the sample data to a file

  BOOL saveData = FALSE;

  // Calculate the number of enabled channels from the channel mask

  int channelCount = 0;
  const int channelsPerBoard = 2;
  for (int channel = 0; channel < channelsPerBoard; channel++) {
    if (channelMask & (1 << channel))
      channelCount++;
  }

  if ((channelCount < 1) || (channelCount == 3) ||
      (channelCount > channelsPerBoard)) {
    printf("Error: Invalid channel mask %08X\n", channelMask);
    return FALSE;
  }

  // Get the sample size in bits, and the on-board memory size in samples per
  // channel

  U8 bitsPerSample;
  U32 maxSamplesPerChannel;
  RETURN_CODE retCode =
      AlazarGetChannelInfo(boardHandle, &maxSamplesPerChannel, &bitsPerSample);
  if (retCode != ApiSuccess) {
    printf("Error: AlazarGetChannelInfo failed -- %s\n",
           AlazarErrorToText(retCode));
    return FALSE;
  }

  // Calculate the size of each DMA transfer in bytes

  U32 bytesPerSample = (bitsPerSample + 7) / 8;
  U32 bytesPerRecord = bytesPerSample * samplesPerRecord;
  U32 samplesPerDmaTransfer =
      samplesPerRecord * recordsPerDmaBuffer * channelCount;
  U32 bytesPerDmaBuffer = bytesPerSample * samplesPerDmaTransfer;

  // Check that the DMA transfer size is a multiple of the page size
  const U32 page_size = 4096;
  if (bytesPerDmaBuffer % page_size) {
    printf("Error: DMA transfer size is not a multiple of the page size\n");
    return FALSE;
  }

  // Check that the DMA transfer size is between 1 and 8 MB.
  if (bytesPerDmaBuffer < (2UL << 20) ||
      bytesPerDmaBuffer > (2UL << 23)) {
    printf("Warning: Size of DMA transfer is not ideal. Best values are between 1 and 8 MB\n");
  }

  // Create a data file if required

  FILE *fpData = NULL;

  if (saveData) {
    fpData = fopen("data.bin", "wb");
    if (fpData == NULL) {
      printf("Error: Unable to create data file -- %u\n", GetLastError());
      return FALSE;
    }
  }

  // Allocate a large user-space buffer for the acquistion data

  U32 dmaTransfersPerUserBuffer = recordsPerUserBuffer / recordsPerDmaBuffer;
  size_t bytesPerUserBuffer =
      (U64)bytesPerDmaBuffer * (U64)dmaTransfersPerUserBuffer;

  const U32 maxDmaBufferCount = 4; // Maximum DMA buffers posted to the board simultaneously
  U32 dmaBufferCount;
  if (dmaTransfersPerUserBuffer > maxDmaBufferCount)
    dmaBufferCount = maxDmaBufferCount;
  else
    dmaBufferCount = dmaTransfersPerUserBuffer;

  printf("User buffer = %lu MB, DMA buffer = %lu x %u KB\n",
         bytesPerUserBuffer >> 20, dmaBufferCount, bytesPerDmaBuffer >> 10);

  BOOL success = TRUE;
#ifdef _WIN32
  Buffer =
      (U16 *)VirtualAlloc(NULL, bytesPerUserBuffer, MEM_COMMIT, PAGE_READWRITE);
#else
  Buffer = (U16 *)valloc(bytesPerUserBuffer);
#endif
  if (Buffer == NULL) {
    printf("Error: Alloc %zd bytes failed\n", bytesPerUserBuffer);
    success = FALSE;
  }

  // Configure the record size

  if (success) {
    retCode =
        AlazarSetRecordSize(boardHandle,       // HANDLE -- board handle
                            0,                 // U32 -- pre-trigger samples
                            samplesPerRecord   // U32 -- post-trigger samples
                            );
    if (retCode != ApiSuccess) {
      printf("Error: AlazarSetRecordSize failed -- %s\n",
             AlazarErrorToText(retCode));
      success = FALSE;
    }
  }

  // Configure the board to make an NPT AutoDMA acquisition

  if (success) {
    U32 recordsPerAcquisition = 0x7fffffff; // Acquire until aborted

    U32 admaFlags =
        ADMA_EXTERNAL_STARTCAPTURE | // Start acquisition when
                                     // AlazarStartCapture is called
        ADMA_NPT | // Acquire multiple records with no-pretrigger samples
        ADMA_INTERLEAVE_SAMPLES;

    retCode = AlazarBeforeAsyncRead(
        boardHandle,              // HANDLE -- board handle
        channelMask,              // U32 -- enabled channel mask
        0,                        // long -- offset from trigger in samples
        samplesPerRecord,         // U32 -- samples per record
        recordsPerDmaBuffer,      // U32 -- records per buffer
        recordsPerAcquisition,    // U32 -- records per acquisition
        admaFlags                 // U32 -- AutoDMA flags
        );
    if (retCode != ApiSuccess) {
      printf("Error: AlazarBeforeAsyncRead failed -- %s\n",
             AlazarErrorToText(retCode));
      success = FALSE;
    }
  }

  // Add the buffers to a list of buffers available to be filled by the board

  U32 dmaTransfersStarted = 0;
  for (U32 bufferIndex = 0; (bufferIndex < dmaBufferCount) && success;
       bufferIndex++) {
    U16 *pBuffer = Buffer + (U64)samplesPerDmaTransfer * dmaTransfersStarted;
    retCode = AlazarPostAsyncBuffer(boardHandle, pBuffer, bytesPerDmaBuffer);
    if (retCode != ApiSuccess) {
      printf("Error: AlazarPostAsyncBuffer %u failed -- %s\n", bufferIndex,
             AlazarErrorToText(retCode));
      success = FALSE;
    }
    dmaTransfersStarted++;
  }

  // Arm the board system to wait for a trigger event to begin the acquisition

  if (success) {
    retCode = AlazarStartCapture(boardHandle);
    if (retCode != ApiSuccess) {
      printf("Error: AlazarStartCapture failed -- %s\n",
             AlazarErrorToText(retCode));
      success = FALSE;
    }
  }

  // Wait for each buffer to be filled, process the buffer, and re-post it to
  // the board.

  if (success) {
    printf("Capturing %d DMA buffers ... press any key to abort\n",
           dmaTransfersPerUserBuffer);

    DWORD startTickCount = GetTickCount();
    U32 dmaTransfersCompleted = 0;
    INT64 bytesTransferred = 0;

    while (dmaTransfersCompleted < dmaTransfersPerUserBuffer) {
      // TODO: Set a buffer timeout that is longer than the time
      //       required to capture all the records in one buffer.

      DWORD timeout_ms = 5000;

      // Wait for the oldest pending dma buffer segment to be filled by the
      // board.

      U16 *pBuffer =
          Buffer + (U64)samplesPerDmaTransfer * dmaTransfersCompleted;
      retCode = AlazarWaitAsyncBufferComplete(boardHandle, pBuffer, timeout_ms);
      if (retCode != ApiSuccess) {
        printf("Error: AlazarWaitAsyncBufferComplete failed -- %s\n",
               AlazarErrorToText(retCode));
        success = FALSE;
      }

      if (success) {
        // The buffer segment is full and has been removed from the list
        // of buffer segments available for the board

        dmaTransfersCompleted++;
        bytesTransferred += bytesPerDmaBuffer;

        // TODO: Process sample data in this buffer segment.

        // NOTE:
        //
        // While you are processing this buffer, the board is already
        // filling the next available buffer(s).
        //
        // You MUST finish processing this buffer and post it back to the
        // board before the board fills all of its available DMA buffers
        // and on-board memory.
        //
        // Records are arranged in the buffer as follows:
        // R0A, R1A, R2A ... RnA, R0B, R1B, R2B ..., R0C, R1C, R2C ..., R0D,
        // R1D, R2D ...
        //
        // Samples values are arranged contiguously in each record.
        // A 14-bit sample code is stored in the most significant bits of
        // each 16-bit sample value.
        //
        // Sample codes are unsigned by default. As a result:
        // - a sample code of 0x0000 represents a negative full scale input
        // signal.
        // - a sample code of 0x2000 represents a ~0V signal.
        // - a sample code of 0x3FFF represents a positive full scale input
        // signal.

        U16 *pRecord = pBuffer;

        for (int channel = 0; (channel < channelCount) && (success == TRUE);
             channel++) {
          for (U32 record = 0;
               (record < recordsPerDmaBuffer) && (success == TRUE);
               record++) {
            if (saveData) {
              // Write record to file

              size_t bytesWritten =
                  fwrite(pRecord, sizeof(BYTE), bytesPerRecord, fpData);
              if (bytesWritten != bytesPerRecord) {
                printf("Error: Write record %u failed -- %u\n", record,
                       GetLastError());
                success = FALSE;
              }
            }

            // Seek to next record in buffer

            pRecord += samplesPerRecord;
          }
        }
      }

      // Add the next buffer segment to the end of the list of available DMA
      // buffers.

      if (success && (dmaTransfersStarted < dmaTransfersPerUserBuffer)) {
        pBuffer = Buffer + (U64)samplesPerDmaTransfer * dmaTransfersStarted;
        retCode =
            AlazarPostAsyncBuffer(boardHandle, pBuffer, bytesPerDmaBuffer);
        if (retCode != ApiSuccess) {
          printf("Error: AlazarPostAsyncBuffer failed -- %s\n",
                 AlazarErrorToText(retCode));
          success = FALSE;
        } else {
          dmaTransfersStarted++;
        }
      }

      // If the acquisition failed, exit the acquisition loop

      if (!success)
        break;

      // If a key was pressed, exit the acquisition loop

      if (_kbhit()) {
        printf("Aborted...\n");
        break;
      }

      // Display progress

      printf("Completed %u buffers\r", dmaTransfersCompleted);
    }

    // Display results

    double transferTime_sec = (GetTickCount() - startTickCount) / 1000.;
    printf("Capture completed in %.2lf sec\n", transferTime_sec);

    U32 recordsTransferred = recordsPerDmaBuffer * dmaTransfersCompleted;

    double buffersPerSec;
    double bytesPerSec;
    double recordsPerSec;

    if (transferTime_sec > 0.) {
      buffersPerSec = dmaTransfersCompleted / transferTime_sec;
      bytesPerSec = bytesTransferred / transferTime_sec;
      recordsPerSec = recordsTransferred / transferTime_sec;

    } else {
      buffersPerSec = 0.;
      bytesPerSec = 0.;
      recordsPerSec = 0.;
    }

    printf("Captured %u DMA buffers (%.4g buffers per sec)\n",
           dmaTransfersCompleted, buffersPerSec);
    printf("Captured %u records (%.4g records per sec)\n", recordsTransferred,
           recordsPerSec);
    printf("Transferred %I64d bytes (%.4g bytes per sec)\n", bytesTransferred,
           bytesPerSec);
  }

  // Abort the acquisition

  retCode = AlazarAbortAsyncRead(boardHandle);
  if (retCode != ApiSuccess) {
    printf("Error: AlazarAbortAsyncRead failed -- %s\n",
           AlazarErrorToText(retCode));
    success = FALSE;
  }

  // Free all memory allocated

  if (Buffer != NULL) {
#ifdef _WIN32
    VirtualFree(Buffer, 0, MEM_RELEASE);
#else
    free(Buffer);
#endif
  }

  // Close the data file

  if (fpData != NULL)
    fclose(fpData);

  return success;
}
